/**
 *  Copyright (C) 2013
 *
 *  RTCFC Implementation
 *
 *  This implementation uses flat combining instead of RCL
 *  (Note: it does not contain the dependency detector server).
 *
 */

#include "../cm.hpp"
#include "algs.hpp"
#include "RedoRAWUtils.hpp"
#include <sys/mman.h>

// Don't just import everything from stm. This helps us find bugs.
using stm::TxThread;
using stm::timestamp;
using stm::WriteSetEntry;
using stm::ValueList;
using stm::ValueListEntry;
using stm::filter_t;
using stm::pad_word_t;

/*
 *  RTCFC definitions
 */

#define PAGE_SIZE	4096
#define CACHE_LINE_SIZE 64
#define r_align(n, r)        (((n) + (r) - 1) & -(r))
#define cache_align(n)       r_align(n , CACHE_LINE_SIZE)
#define pad_to_cache_line(n) (cache_align(n) - (n))

// Server-Client Request states
#define READY 		0
#define PENDING		1
#define ABORTED		2

#define GET_NODES_CMD																										\
		"NODES=/sys/devices/system/node;"																			\
		"CPUS=/sys/devices/system/cpu;"																				\
		"if [ -d $NODES ]; then"																							\
	  "  for F in `ls -d $NODES/node*`; do if [ $(cat $F/cpulist | grep '-') ]; then seq -s ' ' $(cat $F/cpulist | tr -s '-' ' '); else cat $F/cpulist | tr -s ',' ' '; fi; done;" \
		"else"																																\
		"  find $CPUS -maxdepth 1 -name \"cpu[0-9]*\" -exec basename {} \\; | sed -e 's/cpu\\(.*\\)/\\1/g' | tr -s '\n' ' ';" \
		"fi"

/*
 *  RTCFC structures
 */
struct request {
	volatile int state; /* the state of commit request */
	TxThread* tx; 		/* Tx context */
	char pad[pad_to_cache_line(sizeof(int) + sizeof(TxThread*))];
};

/*
 * RTCFC Variables
 */

static struct request* requests = 0; /* requests array */
volatile int rtcfc_num_threads = 0;
volatile int rtcfc_num_cores = 0;

cpu_set_t rtcfc_client_cpuset;
cpu_set_t rtcfc_main_server_cpuset;

volatile bool rtcfc_server_initialized = 0;
volatile int rtcfc_combiner_id = -1;

namespace {

const uintptr_t VALIDATION_FAILED = 1;
NOINLINE uintptr_t validate(TxThread*);
bool irrevoc(STM_IRREVOC_SIG(,));
void onSwitchTo();

template<class CM>
struct RTCFC_Generic {
	static TM_FASTCALL bool begin(TxThread*);
	static TM_FASTCALL void commit(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_ro(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_rw(STM_COMMIT_SIG(,));
	static TM_FASTCALL void* read_ro(STM_READ_SIG(,,));
	static TM_FASTCALL void* read_rw(STM_READ_SIG(,,));
	static TM_FASTCALL void write_ro(STM_WRITE_SIG(,,,));
	static TM_FASTCALL void write_rw(STM_WRITE_SIG(,,,));
	static stm::scope_t* rollback(STM_ROLLBACK_SIG(,,,));
	static void initialize(int id, const char* name);
};

/*
 * RTCFC Methods
 */

void* rcl_anon_mmap(size_t n) {
	void* res = mmap(0, n, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE,
			-1, 0);
	if (res == MAP_FAILED)
		fprintf(stderr, "Cannot allocate anon mmap");
	return res;
}

static void rcl_extract_topology(const char* cmd_nodes) {
	FILE *file;
	char text[1024], *p, *saveptr;
	int nb_nodes = 0;
	int nb_cores = 0;

	/* We use UNIX commands to find the nodes. */
	if (!(file = popen(cmd_nodes, "r")))
		fprintf(stderr, "Cannot open system file");

	/* For each node... */
	while (fgets(text, 1024, file)) {
		for (p = 0, p = strtok_r(text, " ", &saveptr); p;
				p = strtok_r(0, " ", &saveptr))
			nb_cores++;
		nb_nodes++;
	}

	printf("Number of nodes: %d\n", nb_nodes);
	printf("Number of cores: %d\n", nb_cores);

	rtcfc_num_cores = nb_cores;

	// close file
	if (pclose(file) < 0)
		fprintf(stderr, "Cannot close system file");

	// add first core to server cpu-set and all other cores to clients cpu-set
	// These cpu-sets will be used in threads binding
	for (int i = 0; i < nb_cores; i++) {

		// first core is dedicated for server and other cores are dedicated for clients.
		if (i == 0)
			CPU_SET(i, &rtcfc_main_server_cpuset);
		else
			CPU_SET(i, &rtcfc_client_cpuset);
	}
}

void initialize_metadata()
{
	// determine number of cores
	rcl_extract_topology(GET_NODES_CMD);

	// Allocate "requests" and "bf_requests" Arrays
	size_t request_size =
			r_align(sizeof(struct request)*stm::MAX_THREADS, PAGE_SIZE);
	requests = (request*) rcl_anon_mmap(request_size);

	for (unsigned int i = 0; i < stm::MAX_THREADS; i++)
		requests[i].state = READY;

	// mark server as initialized so that meta data is initialized only once
	rtcfc_server_initialized = true;

}

void clear_metadata()
{
	size_t request_size =
			r_align(sizeof(struct request)*stm::MAX_THREADS, PAGE_SIZE);

	// deallocate request array
	munmap(requests,request_size);

	// clear flag when killing server
	rtcfc_server_initialized = false;
}

void* combine(void* dummy) {
//	printf("I'm a combiner!!!!!!!!\n");
	request* req;
	int i = 0;
	int curr_num_threads = rtcfc_num_threads;

	curr_num_threads = rtcfc_num_threads;

	for (i = 0; i < curr_num_threads; i++) {
		req = &(requests[i]);
		// pending transaction commit request
		if (req->state == PENDING)
		{
			// Validation is needed because client may be invalidated after changing its state to PENDING
			if (timestamp.val !=  req->tx->start_time) {
				if ((req->tx->start_time = validate(req->tx)) == VALIDATION_FAILED)
				{
					req->state = ABORTED;
					continue;
				}
			}

			timestamp.val++;

			req->tx->writes.writeback(STM_WHEN_PROTECT_STACK(upper_stack_bound));

			CFENCE;
			timestamp.val++;

			req->state = READY;
		}
	}
	return NULL;
}

/*
 * End of RTCFC server loop
 */

uintptr_t validate(TxThread* tx) {

	while (true) {

		// loop until no commit is executed
        uintptr_t s = timestamp.val;
         if ((s & 1) == 1)
             continue;

		// check the read set
		CFENCE;
		// don't branch in the loop---consider it backoff if we fail
		// validation early
		bool valid = true;
		foreach (ValueList, i, tx->vlist)
			valid &= i->isValid();

		if (!valid)
			return VALIDATION_FAILED;

		// restart if timestamp changed during read set iteration
		CFENCE;
		if (timestamp.val == s)
			return s;
	}
}

bool irrevoc(STM_IRREVOC_SIG(tx,upper_stack_bound)) {

	fprintf(stderr, "Irrevoc is not yet implemented");

	return true;
}

void onSwitchTo() {

	// We just need to be sure that the timestamp is not odd, or else we will
	// block.  For safety, increment the timestamp to make it even, in the event
	// that it is odd.

	if (timestamp.val & 1)
		++timestamp.val;

	// initialize the requests array
	if (!rtcfc_server_initialized)
		initialize_metadata();

	// Create Server thread if not exist...
//	if (!rtcfc_server_initialized) {
//		void* args = (void*) (stm::MAX_THREADS);
//		pthread_t tid;
//		pthread_attr_t attr;
//		pthread_attr_init(&attr);
//		pthread_create(&tid, &attr, &rcl_run_main_server, args);
//
//		// wait until server thread initializes its data-structures
//		while (!rtcfc_server_initialized)
//			;
//	}
}

template<typename CM>
void RTCFC_Generic<CM>::initialize(int id, const char* name) {

	// set the name
	stm::stms[id].name = name;

	// set the pointers
	stm::stms[id].begin = RTCFC_Generic<CM>::begin;
	stm::stms[id].commit = RTCFC_Generic<CM>::commit_ro;
	stm::stms[id].read = RTCFC_Generic<CM>::read_ro;
	stm::stms[id].write = RTCFC_Generic<CM>::write_ro;
	stm::stms[id].irrevoc = irrevoc;
	stm::stms[id].switcher = onSwitchTo;
	stm::stms[id].privatization_safe = true;
	stm::stms[id].rollback = RTCFC_Generic<CM>::rollback;
}

template<class CM>
bool RTCFC_Generic<CM>::begin(TxThread* tx) {

	// bind thread to clinet core-set
	if (!tx->bounded) {
//		pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
//				&rtcfc_client_cpuset);

		//increase number of threads
		//TODO:: Do it more intelligently, and decrease number of threads when thread terminates
		int curr_num_thread = rtcfc_num_threads;
		while (!bcasptr(&rtcfc_num_threads, curr_num_thread,
				curr_num_thread + 1))
			curr_num_thread = rtcfc_num_threads;

		tx->bounded = true;
	}

	// Sample the sequence lock, if it is even decrement by 1
	tx->start_time = timestamp.val & ~(1L);

	// notify the allocator
	tx->allocator.onTxBegin();

	// notify CM
	CM::onBegin(tx);

	return false;
}

template<class CM>
void RTCFC_Generic<CM>::commit(STM_COMMIT_SIG(tx,upper_stack_bound)) {
	fprintf(stderr,
			"COMMIT IS NOT SUPPOSED TO BE CALLED. COMMIT_RO AND COMMIT_RW ARE ONLY USED");
}

template<class CM>
void RTCFC_Generic<CM>::commit_ro(STM_COMMIT_SIG(tx,)) {
	// Since all reads were consistent, and no writes were done, the read-only
	// NOrec transaction just resets itself and is done.
	CM::onCommit(tx);
	tx->vlist.reset();
	OnReadOnlyCommit(tx);

}

template<class CM>
void RTCFC_Generic<CM>::commit_rw(STM_COMMIT_SIG(tx,upper_stack_bound)) {


	// Validation. Saves time of servers on useless invalidated requests
	if (timestamp.val != tx->start_time)
		if ((tx->start_time = validate(tx)) == VALIDATION_FAILED)
			tx->tmabort(tx);

	// send request and then loop
	requests[tx->id - 1].tx = tx;
	requests[tx->id - 1].state = PENDING;
	while (requests[tx->id - 1].state == PENDING)
	{
		if(rtcfc_combiner_id == -1)
		{
			if (bcasptr(&rtcfc_combiner_id, -1, tx->id))
			{
				combine(NULL);
				rtcfc_combiner_id = -1;
			}
		}
		else
			sched_yield();
	}

	if(requests[tx->id - 1].state == ABORTED)
	{
		requests[tx->id - 1].state = READY;
		tx->tmabort(tx);
	}
	else
	{
		// notify CM
		CM::onCommit(tx);

		tx->vlist.reset();
		tx->writes.reset();
		tx->rf->clear();
		tx->wf->clear();

		// This switches the thread back to RO mode.
		OnReadWriteCommit(tx, read_ro, write_ro, commit_ro);
	}
}

template<class CM>
void*
RTCFC_Generic<CM>::read_ro(STM_READ_SIG(tx,addr,mask)) {
	// A read is valid iff it occurs during a period where the seqlock does
	// not change and is even.  This code also polls for new changes that
	// might necessitate a validation.

	// read the location to a temp
	void* tmp = *addr;
	CFENCE;

	// if the timestamp has changed since the last read, we must validate and
	// restart this read
	while (tx->start_time != timestamp.val) {
		if ((tx->start_time = validate(tx)) == VALIDATION_FAILED)
			tx->tmabort(tx);

		tmp = *addr;
		CFENCE;
	}

	// log the address and value
	tx->vlist.insert(STM_VALUE_LIST_ENTRY(addr, tmp, mask));
	tx->rf->add(addr);
	return tmp;
}

template<class CM>
void*
RTCFC_Generic<CM>::read_rw(STM_READ_SIG(tx,addr,mask)) {
	// check the log for a RAW hazard, we expect to miss
	WriteSetEntry log(STM_WRITE_SET_ENTRY(addr, NULL, mask));
	bool found = tx->writes.find(log);
	REDO_RAW_CHECK(found, log, mask);

	// Use the code from the read-only read barrier. This is complicated by
	// the fact that, when we are byte logging, we may have successfully read
	// some bytes from the write log (if we read them all then we wouldn't
	// make it here). In this case, we need to log the mask for the rest of the
	// bytes that we "actually" need, which is computed as bytes in mask but
	// not in log.mask. This is only correct because we know that a failed
	// find also reset the log.mask to 0 (that's part of the find interface).
	void* val = read_ro(tx, addr STM_MASK(mask & ~log.mask));
	REDO_RAW_CLEANUP(val, found, log, mask);
	tx->rf->add(addr);
	return val;
}

template<class CM>
void RTCFC_Generic<CM>::write_ro(STM_WRITE_SIG(tx,addr,val,mask)) {
	// buffer the write, and switch to a writing context
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	tx->wf->add(addr);
	OnFirstWrite(tx, read_rw, write_rw, commit_rw);
}

template<class CM>
void RTCFC_Generic<CM>::write_rw(STM_WRITE_SIG(tx,addr,val,mask)) {
	// just buffer the write
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	tx->wf->add(addr);
}

template<class CM>
stm::scope_t*
RTCFC_Generic<CM>::rollback(
		STM_ROLLBACK_SIG(tx, upper_stack_bound, except, len)) {

	stm::PreRollback(tx);

	// notify CM
	CM::onAbort(tx);

	// Perform writes to the exception object if there were any... taking the
	// branch overhead without concern because we're not worried about
	// rollback overheads.
	STM_ROLLBACK(tx->writes, upper_stack_bound, except, len);

	tx->vlist.reset();
	tx->writes.reset();
	return stm::PostRollback(tx, read_ro, write_ro, commit_ro);

}
} // (anonymous namespace)

// Register RTCFC initializer functions. Do this as declaratively as
// possible. Remember that they need to be in the stm:: namespace.
#define FOREACH_RTCFC(MACRO)                    \
    MACRO(RTCFC, HyperAggressiveCM)             \
    MACRO(RTCFCHour, HourglassCM)               \
    MACRO(RTCFCBackoff, BackoffCM)              \
    MACRO(RTCFCHB, HourglassBackoffCM)

#define INIT_RTCFC(ID, CM)                      \
    template <>                                 \
    void initTM<ID>() {                         \
        RTCFC_Generic<CM>::initialize(ID, #ID);     \
    }

namespace stm {
FOREACH_RTCFC(INIT_RTCFC)
}

#undef FOREACH_RTCFC
#undef INIT_RTCFC
